@@ -4,6 +4,7 @@
 #include <glib.h>
 
 #include <stdlib.h>
+#include <stdio.h>
 #include <cups/cups.h>
 #include <cups/language.h>
 #include <cups/http.h>
@@ -29,7 +30,7 @@ typedef struct
 
 typedef struct
 {
-	gboolean cancelled;
+	gboolean canceled;
 	gboolean direct_callback;
 	guint id;
 	GnomeCupsConnection *connection;
@@ -42,12 +43,19 @@ typedef struct
 
 	ipp_t *request;
 	char *path;
+	int output_fd;
 } GnomeCupsRequest;
 
 static void request_thread_main (GnomeCupsRequest *request, gpointer unused);
+static ipp_t * gnome_cups_request_execute_sync_internal (ipp_t *request,
+							 const char *server,
+							 const char *path,
+							 int output_fd,
+							 GError **error);
 static guint gnome_cups_request_execute_async_internal (ipp_t      *request, 
 							const char *server, 
 							const char *path, 
+							int output_fd,
 							gboolean direct_callback,
 							GnomeCupsAsyncRequestCallback callback,
 							gpointer cb_data,
@@ -109,7 +117,7 @@ get_error_string (ipp_status_t error)
 			N_("Temporary error"),
 			N_("The printer is not accepting jobs"),
 			N_("The printer is busy"),
-			N_("The job has been cancelled"),
+			N_("The job has been canceled"),
 			N_("Multiple-document jobs are not supported"),
 			N_("The printer is deactivated"),
 		};                                                                             
@@ -224,13 +232,13 @@ idle_stop_unused_threads (gpointer unused)
 		g_static_mutex_unlock (&request_mutex);
 		return FALSE;
 	}
+	g_static_mutex_unlock (&request_mutex);
 
 	g_thread_pool_stop_unused_threads ();
 
-	g_static_mutex_unlock (&request_mutex);
 	return TRUE;
 }
-
+ 
 static gboolean
 close_unused_connection (const char *server,
 			 GnomeCupsConnection *connection,
@@ -238,7 +246,8 @@ close_unused_connection (const char *server,
 {
 	gboolean ret;
 
-	g_mutex_lock (connection->mutex);
+	if (!g_mutex_trylock (connection->mutex))
+		return FALSE;
 	ret = (g_atomic_int_get (&connection->refcount) == 0
 	       && (current_time->tv_sec - connection->use_time.tv_sec > 30));
 	g_mutex_unlock (connection->mutex);
@@ -283,6 +292,9 @@ static void
 gnome_cups_request_connection_destroy (GnomeCupsConnection *connection)
 {
 	g_mutex_lock (connection->mutex);
+	/* FIXME - we shouldn't do this in the main thread
+	 * probably need a cleanup thread to do this
+	 */
 	if (connection->http)
 		httpClose (connection->http);
 	g_free (connection->server);
@@ -294,14 +306,15 @@ gnome_cups_request_connection_destroy (GnomeCupsConnection *connection)
 static gboolean
 idle_signal_request_complete (GnomeCupsRequest *request)
 {
-	if (!request->cancelled && request->callback)
+	if (!request->canceled && request->callback) {
 		request->callback (request->id,
 				   request->path,
 				   request->response,
 				   request->error,
 				   request->cb_data);
-	else {
-		ippDelete (request->response);
+	} else {
+		if (request->response)
+			ippDelete (request->response);
 	}
 
 	g_static_mutex_lock (&request_mutex);
@@ -325,17 +338,53 @@ do_signal_complete (GnomeCupsRequest *request)
 		g_idle_add ((GSourceFunc) idle_signal_request_complete, request);
 }
 
+static void
+dump_request (ipp_t *req)
+{
+	unsigned i;
+	ipp_attribute_t *attr;
+
+	for (attr = req->attrs; attr != NULL; attr = attr->next) {
+		g_print ("%s", attr->name);
+		for (i = 0 ; i < attr->num_values ; i++) {
+			g_print ("\t[%d] = ", i);
+			switch (attr->value_tag) {
+			case IPP_TAG_INTEGER:
+			case IPP_TAG_ENUM:
+				g_print ("%d\n", attr->values[i].integer);
+				break;
+
+			case IPP_TAG_STRING:
+			case IPP_TAG_TEXT:
+			case IPP_TAG_NAME:
+			case IPP_TAG_KEYWORD:
+			case IPP_TAG_URI:
+			case IPP_TAG_CHARSET:
+			case IPP_TAG_LANGUAGE:
+			case IPP_TAG_MIMETYPE:
+				g_print ("'%s'\n", attr->values[i].string.text);
+				break;
+
+			case IPP_TAG_BOOLEAN:
+				g_print ("%d\n", (int)attr->values[i].boolean);
+				break;
+
+			default:
+				g_print ("unprintable\n");
+			}
+		}
+	}
+}
+
 static void
 request_thread_main (GnomeCupsRequest *request,
 		     G_GNUC_UNUSED gpointer unused)
 {
-	ipp_t *response;
 	ipp_status_t status;
-	unsigned i;
 
 	g_return_if_fail (request != NULL);
 
-	if (request->cancelled) {
+	if (request->canceled) {
 		do_signal_complete (request);
 		return;
 	}
@@ -346,13 +395,8 @@ request_thread_main (GnomeCupsRequest *request,
 	g_mutex_lock (request->connection->mutex);
 	
 	if (_gnome_cups_debug && request->request != NULL) {
-		ipp_attribute_t *attr;
-		for (attr = request->request->attrs; attr != NULL; attr = attr->next) {
-			g_print ("%s", attr->name);
-			for (i = 0 ; i < attr->num_values ; i++)
-				if (attr->value_tag >= IPP_TAG_STRING)
-				    g_print ("\t[%d] = '%s'\n", i, attr->values[i].string.text);
-		}
+		g_print ("request = \n");
+		dump_request (request->request);
 	}
 
 	g_get_current_time (&request->connection->use_time);
@@ -361,16 +405,24 @@ request_thread_main (GnomeCupsRequest *request,
 	if (!request->connection->http)
 		request->connection->http = httpConnectEncrypt (request->connection->server, ippPort(), cupsEncryption());
 
-	response = cupsDoRequest (request->connection->http, request->request,
-				  request->path);
+	if (request->request) {
+		request->response = cupsDoRequest (request->connection->http,
+			request->request, request->path);
 
-	/* FIXME - not currently threadsafe, but cups returns NULL on
-	 * any error.  Thus we just set the status to an internal error
-	 * for now.
-	 */
-	status = cupsLastError ();
-	if (response == NULL)
+		/* FIXME - not currently threadsafe, but cups returns NULL on
+		 * any error.  Thus we just set the status to an internal error
+		 * for now.
+		 */
+		status = cupsLastError ();
+		if (request->response == NULL)
+			status = IPP_INTERNAL_ERROR;
+	} else if (request->output_fd >= 0) {
+		status = cupsGetFd(request->connection->http, request->path, request->output_fd);
+		request->response = NULL;
+	} else {
 		status = IPP_INTERNAL_ERROR;
+		g_warning ("Neither request nor output_fd set\n");
+	}
 
 	g_atomic_int_dec_and_test (&request->connection->refcount);
 	g_mutex_unlock (request->connection->mutex);
@@ -378,13 +430,17 @@ request_thread_main (GnomeCupsRequest *request,
 		g_print ("<<<<----- unlocking %p\n", request->connection);
 	}
 
-	if (status > IPP_OK_CONFLICT && request->error != NULL) {
-		*(request->error) = g_error_new (GNOME_CUPS_ERROR, 
-						 status,
-						 get_error_string (status));
+	if (status > IPP_OK_CONFLICT) {
+		g_warning ("failed request with status %d", status);
+		if (request->error != NULL)
+			*(request->error) = g_error_new (GNOME_CUPS_ERROR, 
+							 status,
+							 get_error_string (status));
+	} else if (request->response && _gnome_cups_debug) {
+		g_print ("response = \n");
+		dump_request (request->response);
 	}
 
-	request->response = response;
 	do_signal_complete (request);
 	
 	return;
@@ -518,6 +574,33 @@ async_wrapper_cb (guint id, const char *path,
 
 ipp_t *
 gnome_cups_request_execute (ipp_t *request, const char *server, const char *path, GError **err)
+{
+	return gnome_cups_request_execute_sync_internal (request, server, path, -1, err);
+}
+
+/**
+ * gnome_cups_request_file:
+ * @server: The hostname of the IPP server to connect to
+ * @path: The URI of the file to retrieve 
+ * @output_fd: The file descriptor to which to write the retrieved file
+ * @err: A #GError for storing errors
+ *
+ * Retrieve the file named by @path via IPP, writing * the data to @output_fd.
+ **/
+void
+gnome_cups_request_file (const char *server, const char *path,
+			 int output_fd,
+			 GError **err)
+{
+	(void) gnome_cups_request_execute_sync_internal (NULL, server, path, output_fd, err);
+}
+
+static ipp_t *
+gnome_cups_request_execute_sync_internal (ipp_t *request,
+					  const char *server,
+					  const char *path,
+					  int output_fd,
+					  GError **err)
 {
 	guint id;
 	GnomeCupsAsyncWrapperData data;
@@ -529,7 +612,7 @@ gnome_cups_request_execute (ipp_t *request, const char *server, const char *path
 	data.error = err;
 
 	id = gnome_cups_request_execute_async_internal (request, server, path,
-							TRUE,
+							output_fd, TRUE,
 							async_wrapper_cb,
 							&data,
 							NULL);
@@ -542,7 +625,7 @@ gnome_cups_request_execute (ipp_t *request, const char *server, const char *path
 
 	g_mutex_free (data.mutex);
 	g_cond_free (data.cond);
-	
+
 	return data.response;
 }
 
@@ -568,15 +651,44 @@ gnome_cups_request_execute_async (ipp_t      *request,
 				  GDestroyNotify destroy_notify)
 {
 	return gnome_cups_request_execute_async_internal (request, server,
-							  path, FALSE,
-							  callback, cb_data,
-							  destroy_notify);
+							  path, -1,
+							  FALSE, callback,
+							  cb_data, destroy_notify);
+}
+
+/**
+ * gnome_cups_request_file_async:
+ * @server: The hostname of the IPP server to connect to
+ * @path: The URI of the file to retrieve 
+ * @output_fd: The file descriptor to which to write the retrieved file
+ * @callback: A #GnomeCupsAsyncRequestCallback.
+ * @cb_data: Data for the callback
+ * @destroy_notify: A function to free the callback data
+ * @returns: an operation ID, suitable for passing to gnome_cups_request_cancel
+ *
+ * Creates a new asynchronous IPP request to retrieve the specified file, which will
+ * invoke @cb_data when complete.  Note that the "response" parameter of the callback
+ * will always be #NULL.
+ **/
+guint
+gnome_cups_request_file_async (const char *server, 
+			       const char *path, 
+			       int output_fd,
+			       GnomeCupsAsyncRequestCallback callback,
+			       gpointer cb_data,
+			       GDestroyNotify destroy_notify)
+{
+	return gnome_cups_request_execute_async_internal (NULL, server,
+							  path, output_fd,
+							  FALSE, callback,
+							  cb_data, destroy_notify);
 }
 
 static guint
 gnome_cups_request_execute_async_internal (ipp_t      *request, 
 					   const char *server, 
 					   const char *path, 
+					   int output_fd, 
 					   gboolean direct_callback,
 					   GnomeCupsAsyncRequestCallback callback,
 					   gpointer cb_data,
@@ -609,12 +721,13 @@ gnome_cups_request_execute_async_internal (ipp_t      *request,
 	
 	req = g_new0 (GnomeCupsRequest, 1);
 	req->connection = connection;
-	req->cancelled = FALSE;
+	req->canceled = FALSE;
 	req->request = request;
 	req->callback = callback;
 	req->cb_data = cb_data;
 	req->destroy_notify = destroy_notify;
 	req->path = g_strdup (path);
+	req->output_fd = output_fd;
 	req->direct_callback = direct_callback;
 	req->error = NULL;
 
@@ -635,7 +748,7 @@ gnome_cups_request_cancel (guint request_id)
 	
 	g_static_mutex_lock (&request_mutex);
 	if ((request = g_hash_table_lookup (request_map, GUINT_TO_POINTER (request_id))) != NULL) {
-		request->cancelled = TRUE;
+		request->canceled = TRUE;
 	}
 	g_static_mutex_unlock (&request_mutex);
 }
